#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include <libusb.h>
#include "crc.h"

/* platform crap */
#if defined(__APPLE__)
#  include <libkern/OSByteOrder.h>
#  define htole32(x) OSSwapHostToLittleInt32((x))
#  define htobe32(x) OSSwapHostToBigInt32((x))
#  define be32toh(x) OSSwapBigToHostInt32((x))
#elif defined(_WIN32)
#  include <stdlib.h>
#  include <winsock2.h>
#  if BYTE_ORDER == LITTLE_ENDIAN
#    define htole32(x) (x)
#    define htobe32(x) _byteswap_ulong((x))
#    define be32toh(x) _byteswap_ulong((x))
#  else
#    define htole32(x) _byteswap_ulong((x))
#    define htobe32(x) (x)
#    define be32toh(x) (x)
#  endif
#else
#  define _BSD_SOURCE
#  include <endian.h>
#endif

#if defined(__GNUC__) || (defined(__has_attribute) && __has_attribute(packed))
#  define packed(x)         x __attribute__((packed))
#else
#  define packed(x)         __pragma(pack(push, 1)) x __pragma(pack(pop))
#endif


/* helpers */

static const char *progname = NULL;

#define warn(msg, ...) fprintf(stderr, "%s: %s: " msg "\n", progname, __func__ __VA_OPT__(,) __VA_ARGS__)


/* ST recovery mode defines */

#define ST_USB_VID           0x0483
#define ST_USB_PID_SPEAR300  0x3801
#define ST_USB_PID_SPEAR600  0x7261
#define ST_USB_PID_SPEAR1340 0x3802

#define ST_USB_TIMEOUT         10000
#define ST_USB_CONFIG_ID       1
#define ST_USB_INTERFACE_ID    0
#define ST_USB_IN_ENDPOINT_ID  1
#define ST_USB_OUT_ENDPOINT_ID 2

#define ST_PAYLOAD_BLOCKSIZE   1024
#define ST_PAYLOAD_DDR_DRIVER_NAME    "XLOADER"
#define ST_PAYLOAD_FIRMWARE_NAME      "U-BOOT"
#define ST_PAYLOAD_FIRMWARE_NAME_ALT  "UBOOT"

enum st_payload_type {
	ST_PAYLOAD_DDR_DRIVER = 0x11,
	ST_PAYLOAD_FIRMWARE   = 0x22,
};

/* all fields little-endian */
packed(struct st_payload_desc {
	uint32_t type;
	uint32_t size;
	uint32_t addr;
});

/* this is just the uImage legacy header; below defnitions taken from https://source.denx.de/u-boot/u-boot/-/blob/v2023.10/include/image.h */
#define IH_MAGIC	0x27051956	/* Image Magic Number		*/
#define IH_NMLEN		32	/* Image Name Length		*/

packed(struct legacy_img_hdr {
	uint32_t ih_magic;           /* Image Header Magic Number */
	uint32_t ih_hcrc;            /* Image Header CRC Checksum */
	uint32_t ih_time;            /* Image Creation Timestamp */
	uint32_t ih_size;            /* Image Data Size */
	uint32_t ih_load;            /* Data Load Address*/
	uint32_t ih_ep;              /* Entry Point Address */
	uint32_t ih_dcrc;            /* Image Data CRC Checksum */
	uint8_t  ih_os;              /* Operating System */
	uint8_t  ih_arch;            /* CPU architecture */
	uint8_t  ih_type;            /* Image Type */
	uint8_t  ih_comp;            /* Compression Type */
	uint8_t  ih_name[IH_NMLEN];  /* Image Name */
});

packed(struct st_payload_header {
	struct legacy_img_hdr uimage;
	char data[0];
});


/* forward declarations */

struct st_device_context;
struct st_device;
static int st_dev_init(struct st_device *dev, struct st_device_context *ctx, libusb_device_handle *usb_handle);


/* recovery device enumeration logic */

enum st_spear_model {
	ST_SPEAR_NONE,
	ST_SPEAR300,
	ST_SPEAR600,
	ST_SPEAR1340,
};

struct st_device_context {
	libusb_context      *usb_ctx;
	enum st_spear_model  model;
};

static int st_is_init(struct st_device_context *ctx)
{
	return ctx->usb_ctx != NULL;
}

static int st_init(struct st_device_context *ctx, enum st_spear_model model)
{
	int r = libusb_init(&ctx->usb_ctx);
	if (r < 0) {
		warn("libusb_init() failed: %s", libusb_strerror(r));
		return -1;
	}
	ctx->model = model;
	return 0;
}

static void st_exit(struct st_device_context *ctx)
{
	if (st_is_init(ctx)) {
		libusb_exit(ctx->usb_ctx);
		ctx->usb_ctx = NULL;
	}
}
struct st_device_filter {
	uint16_t usb_vid;
	uint16_t usb_pid;
};

static int st_find(struct st_device_context *ctx, struct st_device *dev, struct st_device_filter *filter)
{
	if (!st_is_init(ctx)) {
		warn("called without being initialised (did you call st_init()?)");
		return -1;
	}

	if (ctx->model != ST_SPEAR_NONE) {
		if (!filter->usb_vid)
			filter->usb_vid = ST_USB_VID;
		if (!filter->usb_pid) {
			switch (ctx->model) {
			case ST_SPEAR300:;
				filter->usb_pid = ST_USB_PID_SPEAR300;
				break;
			case ST_SPEAR600:;
				filter->usb_pid = ST_USB_PID_SPEAR600;
				break;
			case ST_SPEAR1340:;
				filter->usb_pid = ST_USB_PID_SPEAR1340;
				break;
			default:
				break;
			}
		}
	}

	libusb_device **devs;
	ssize_t ndevs = libusb_get_device_list(ctx->usb_ctx, &devs);
	if (ndevs < 0) {
		warn("libusb_get_device_list() failed: %s", libusb_strerror(ndevs));
		return -1;
	}

	libusb_device_handle *chosen = NULL;
	for (ssize_t i = 0; i < ndevs; i++) {
		struct libusb_device_descriptor desc;
		int r = libusb_get_device_descriptor(devs[i], &desc);
		if (r < 0) {
			warn("libusb_get_device_descriptor() for device #%zu failed: %s", i, libusb_strerror(r));
			continue;
		}
		if (filter->usb_vid && desc.idVendor != filter->usb_vid)
			continue;
		if (filter->usb_pid && desc.idProduct != filter->usb_pid)
			continue;

		r = libusb_open(devs[i], &chosen);
		if (r < 0) {
			warn("libusb_open() for device #%zu failed: %s", i, libusb_strerror(r));
			chosen = NULL;
			continue;
		}
		break;
	}
	/* libusb_open() refs our chosen device, so we can unref all of them safely here */
	libusb_free_device_list(devs, 1);

	if (!chosen) {
		warn("could not find suitable USB device for VID = %04x, PID = %04x", filter->usb_vid, filter->usb_pid);
		return -1;
	}
	return st_dev_init(dev, ctx, chosen);
}


/* recovery device logic */

struct st_device {
	struct st_device_context *ctx;
	libusb_device_handle     *usb_handle;
	int                       usb_claimed_interface;
};

static int st_dev_is_init(const struct st_device *dev)
{
	return dev->ctx && dev->usb_handle;
}

static int st_dev_is_open(const struct st_device *dev)
{
	return st_dev_is_init(dev) && dev->usb_claimed_interface != -1;
}

static int st_dev_open(struct st_device *dev)
{
	if (!st_dev_is_init(dev)) {
		warn("called without being initialised");
		return -1;
	}
	if (st_dev_is_open(dev)) {
		warn("called while already open");
		return -1;
	}
	(void)libusb_set_auto_detach_kernel_driver(dev->usb_handle, 1);

	int r = libusb_set_configuration(dev->usb_handle, ST_USB_CONFIG_ID);
	if (r < 0) {
		warn("libusb_set_configuration() failed: %s", libusb_strerror(r));
		return -1;
	}
	r = libusb_claim_interface(dev->usb_handle, ST_USB_INTERFACE_ID);
	if (r < 0) {
		warn("libusb_claim_interface() failed: %s", libusb_strerror(r));
		return -1;
	}
	dev->usb_claimed_interface = ST_USB_INTERFACE_ID;
	return 0;
}

static int st_dev_close(struct st_device *dev)
{
	if (!st_dev_is_open(dev)) {
		warn("called without being open");
		return -1;
	}
	int r = libusb_release_interface(dev->usb_handle, dev->usb_claimed_interface);
	if (r < 0) {
		warn("libusb_release_interface() failed: %s", libusb_strerror(r));
		return -1;
	}
	dev->usb_claimed_interface = -1;
	return 0;
}

static int st_dev_init(struct st_device *dev, struct st_device_context *ctx, libusb_device_handle *usb_handle)
{
	dev->ctx = ctx;
	dev->usb_handle = usb_handle;
	dev->usb_claimed_interface = -1;
	return 0;
}

static void st_dev_exit(struct st_device *dev)
{
	if (dev->usb_handle) {
		if (dev->usb_claimed_interface != -1) {
			(void)st_dev_close(dev);
			/* in case the above failed */
			dev->usb_claimed_interface = -1;
		}
		libusb_close(dev->usb_handle);
		dev->usb_handle = NULL;
	}
	dev->ctx = NULL;
}

#include <unistd.h>
static int st_dev_send(struct st_device *dev, const void *buf, size_t sz)
{
	if (!st_dev_is_open(dev)) {
		warn("called without being open");
		return -1;
	}

	size_t total = 0;
	const char *p = buf;
	while (total < sz) {
		size_t chunksize = sz - total;
		if (chunksize > ST_PAYLOAD_BLOCKSIZE)
			chunksize = ST_PAYLOAD_BLOCKSIZE;
		int transferred;
		int r = libusb_bulk_transfer(dev->usb_handle, ST_USB_OUT_ENDPOINT_ID, (unsigned char *)p, chunksize, &transferred, ST_USB_TIMEOUT);
		if (r < 0) {
			putc('x', stderr);
			putc('\n', stderr);
			warn("send failed at position %zu: %s", total, libusb_strerror(r));
			return -2;
		}
		putc('.', stderr);
		total += transferred;
		p += transferred;
	}
	putc('\n', stderr);
	return 0;
}

static const struct st_payload_header *st_payload_verify(enum st_payload_type type, const void *buf, size_t sz, size_t *outsz)
{
	if (outsz)
		*outsz = 0;

	const struct st_payload_header *hdr = buf;
	if (sz < sizeof(*hdr)) {
		warn("bad header: size < %zu", sizeof(*hdr));
		return NULL;
	}

	/* verify header */
	const struct legacy_img_hdr *uimage = &hdr->uimage;
	if (be32toh(uimage->ih_magic) != IH_MAGIC) {
		warn("bad header: got magic %x, expected %x", uimage->ih_magic, IH_MAGIC);
		return NULL;
	}
	size_t datasz = be32toh(uimage->ih_size);
	if (sz < datasz + sizeof(*hdr)) {
		warn("bad header: missing data (%zu < %zu)", sz, datasz + sizeof(*hdr));
		return NULL;
	}
	uint32_t header_crc = be32toh(uimage->ih_hcrc);
	struct legacy_img_hdr copy;
	memcpy(&copy, uimage, sizeof(copy));
	copy.ih_hcrc = htobe32(0);
	uint32_t header_calc_crc = crc32(&copy, sizeof(copy));
	if (header_calc_crc != header_crc) {
		warn("bad header: got checksum %x, expected %x", header_calc_crc, header_crc);
		return NULL;
	}

	/* verify data */
	uint32_t data_crc = be32toh(uimage->ih_dcrc);
	uint32_t data_calc_crc = crc32(&hdr->data, datasz);
	if (data_calc_crc != data_crc) {
		warn("bad header: got data checksum %x, expected %x", data_calc_crc, data_crc);
		return NULL;
	}

	/* verify name */
	const char *image_name = (const char *)uimage->ih_name;
	switch (type) {
	case ST_PAYLOAD_DDR_DRIVER:
		if (!strncasecmp(image_name, ST_PAYLOAD_DDR_DRIVER_NAME, sizeof(uimage->ih_name)))
			break;
		warn("bad header: got name %*s, expected %s", (int)sizeof(uimage->ih_name), image_name, ST_PAYLOAD_DDR_DRIVER_NAME);
		return NULL;
	case ST_PAYLOAD_FIRMWARE:
		if (!strncasecmp(image_name, ST_PAYLOAD_FIRMWARE_NAME, sizeof(uimage->ih_name)))
			break;
		if (!strncasecmp(image_name, ST_PAYLOAD_FIRMWARE_NAME_ALT, sizeof(uimage->ih_name)))
			break;
		warn("bad header: got name %*s, expected %s or %s", (int)sizeof(uimage->ih_name), image_name,
			ST_PAYLOAD_FIRMWARE_NAME, ST_PAYLOAD_FIRMWARE_NAME_ALT);
		return NULL;
	default:
		warn("bad type: %d", type);
		return NULL;
	}

	if (outsz)
		*outsz = datasz;
	return hdr;
}

static int st_dev_send_payload(struct st_device *dev, enum st_payload_type type, void *buf, size_t sz)
{
	if (!st_dev_is_open(dev)) {
		warn("called without being open");
		return -1;
	}

	size_t datasz;
	const struct st_payload_header *payload = st_payload_verify(type, buf, sz, &datasz);
	if (!payload)
		return -2;

	/* build and transfer payload descriptor */
	struct st_payload_desc desc;
	desc.type = htole32((uint32_t)type);
	desc.addr = htole32(be32toh(payload->uimage.ih_load));
	desc.size = htole32(datasz);

	int r = st_dev_send(dev, &desc, sizeof(desc));
	if (r < 0) {
		warn("payload descriptor transfer failed");
		return -3;
	}

	/* transfer payload */
	r = st_dev_send(dev, &payload->data, datasz);
	if (r < 0) {
		warn("payload transfer failed");
		return -4;
	}

	return 0;
}


/* CLI helpers */

#include <stdlib.h>
#include <limits.h>
#include <errno.h>
#include <getopt.h>

static int parse_uint(const char *s, unsigned long *out)
{
	char *p;
	errno = 0;
	unsigned long l = strtoul(s, &p, 0);
	if (*p)
		return 0;
	if (l == 0 && errno)
		return 0;
	if (l == ULONG_MAX && errno == ERANGE)
		return 0;
	*out = l;
	return 1;
}

static void *read_file(const char *s, size_t *sz)
{
	*sz = 0;
	FILE *f = fopen(s, "rb");
	if (!f) {
		warn("%s: could not open: %s", s, strerror(errno));
		return NULL;
	}
	if (fseek(f, 0, SEEK_END) < 0) {
		warn("%s: could not seek to end: %s", s, strerror(errno));
		fclose(f);
		return NULL;
	}
	long size = ftell(f);
	if (fseek(f, 0, SEEK_SET) < 0) {
		warn("%s: could not seek to start: %s", s, strerror(errno));
		fclose(f);
		return NULL;
	}
	void *p = malloc(size);
	if (!p) {
		warn("could not allocate %lu bytes: %s", size, strerror(errno));
		fclose(f);
		return NULL;
	}

	long total = 0;
	while (total < size) {
		size_t r = fread((char *)p + total, 1, total - size, f);
		total += r;
		if (r == 0) {
			if (ferror(f)) {
				warn("%s: could not read: %s", s, strerror(errno));
				total = -1;
				break;
			}
			if (feof(f) && total < size) {
				warn("%s: file changed size while being read (%lu < %lu)", s, total, size);
				total = -1;
				break;
			}
		}
	}
	(void)fclose(f);

	if (total < 0) {
		free(p);
		p = NULL;
	} else {
		*sz = size;
	}
	return p;
}


/* CLI glue */

static const struct option opts[] = {
	{ "help",        no_argument,       NULL, 'h' },
	{ "device-type", required_argument, NULL, 't' },
	{ "device-vid",  required_argument, NULL, 'v' },
	{ "device-pid",  required_argument, NULL, 'p' },
	{ 0 },
};

static void usage(const char *progname)
{
	fprintf(stderr, "usage: %s -t TYPE [-v VID] [-p PID] DDRDRIVER FIRMWARE\n", progname);
}

int main(int argc, char **argv)
{
	progname = strrchr(argv[0], '/');
	progname = progname ? progname + 1 : argv[0];

	int opt;
	enum st_spear_model model = ST_SPEAR_NONE;
	struct st_device_filter filter = { 0 };

	while ((opt = getopt_long(argc, argv, "ht:v:p:", opts, NULL)) != -1) {
		switch (opt) {
		case 'h':
			usage(progname);
			return 0;
		case 't':
			if (!strcasecmp(optarg, "spear300") || !strcasecmp(optarg, "300"))
				model = ST_SPEAR300;
			else if (!strcasecmp(optarg, "spear600") || !strcasecmp(optarg, "600"))
				model = ST_SPEAR600;
			else if (!strcasecmp(optarg, "spear1300") || !strcasecmp(optarg, "1300"))
				model = ST_SPEAR1340;
			else if (!strcasecmp(optarg, "spear1340") || !strcasecmp(optarg, "1340"))
				model = ST_SPEAR1340;
			else {
				fprintf(stderr, "%s: -t/--device-type: unknown device type: %s (known: spear300, spear600, spear1340)\n", progname, optarg);
				return 255;
			}
			break;
		case 'v':;
			unsigned long vid;
			if (!parse_uint(optarg, &vid) || vid >= 0x10000) {
				fprintf(stderr, "%s: -v/--device-vid: invalid VID: %s\n", progname, optarg);
				return 255;
			}
			filter.usb_vid = vid;
			break;
		case 'p':;
			unsigned long pid;
			if (!parse_uint(optarg, &pid) || pid >= 0x10000) {
				fprintf(stderr, "%s: -p/--device-pid: invalid PID: %s\n", progname, optarg);
				return 255;
			}
			filter.usb_pid = pid;
			break;
		default:
			fprintf(stderr, "%s: -%c: unknown option\n", progname, opt);
			usage(progname);
			return 255;
		}
	}

	if (model == ST_SPEAR_NONE) {
		fprintf(stderr, "%s: -t/--device-type not specified\n", progname);
		usage(progname);
		return 255;
	}
	if (argc - optind < 2) {
		fprintf(stderr, "%s: %sFIRMWARE not specified\n", progname, argc == optind ? "DDRDRIVER and " : "");
		usage(progname);
		return 255;
	}

	int r = 0;
	void *driver = NULL, *firmware = NULL;
	struct st_device_context ctx = { 0 };
	struct st_device dev = { 0 };

	size_t driver_size;
	const char *driver_file = argv[optind];
	driver = read_file(driver_file, &driver_size);
	if (!driver) {
		fprintf(stderr, "%s: could not read DDR driver file\n", progname);
		r = 255;
		goto out;
	}
	if (!st_payload_verify(ST_PAYLOAD_DDR_DRIVER, driver, driver_size, NULL)) {
		fprintf(stderr, "%s: DDR driver verification failed\n", progname);
		r = 5;
		goto out;
	}

	size_t firmware_size;
	const char *firmware_file = argv[optind + 1];
	firmware = read_file(firmware_file, &firmware_size);
	if (!firmware) {
		fprintf(stderr, "%s: could not read firmware file\n", progname);
		r = 255;
		goto out;
	}
	if (!st_payload_verify(ST_PAYLOAD_FIRMWARE, firmware, firmware_size, NULL)) {
		fprintf(stderr, "%s: firmware verification failed\n", progname);
		r = 6;
		goto out;
	}

	fprintf(stderr, "%s: opening device...\n", progname);
	if (st_init(&ctx, model) < 0) {
		fprintf(stderr, "%s: could not initialize device context\n", progname);
		r = 1;
		goto out;
	}
	if (st_find(&ctx, &dev, &filter) < 0) {
		fprintf(stderr, "%s: device not found, is it powered on and in recovery?\n", progname);
		r = 2;
		goto out;
	}
	if (st_dev_open(&dev) < 0) {
		fprintf(stderr, "%s: could not claim device\n", progname);
		r = 2;
		goto out;
	}

	fprintf(stderr, "%s: sending DDR driver %s...\n", progname, driver_file);
	if (st_dev_send_payload(&dev, ST_PAYLOAD_DDR_DRIVER, driver, driver_size) < 0) {
		fprintf(stderr, "%s: could not send DDR driver\n", progname);
		r = 3;
		goto out;
	}

	fprintf(stderr, "%s: sending firmware %s...\n", progname, firmware_file);
	if (st_dev_send_payload(&dev, ST_PAYLOAD_FIRMWARE, firmware, firmware_size) < 0) {
		fprintf(stderr, "%s: could not send firmware\n", progname);
		r = 4;
		goto out;
	}

out:
	if (firmware)
		free(firmware);
	if (driver)
		free(driver);
	if (st_dev_is_open(&dev))
		st_dev_close(&dev);
	if (st_dev_is_init(&dev))
		st_dev_exit(&dev);
	if (st_is_init(&ctx))
		st_exit(&ctx);

	return r;
}
